import cv2
import numpy as np

cv_vers = int(cv2.__version__[0])


def _channel_check(img, type_):
    """
    Convenience function for raising an exception if the input image is not
    of the correct format.

    """
    exc_info = {2: "grayscale", 3: "BGR"}

    channels = len(np.asarray(img).shape)
    if channels != type_:
        raise ValueError(
            f"Input image needs to be {exc_info[type_]} or have {type_} channels. Instead an image with {channels} channels was provided."
        )


def find_contours(img, *args, **kwargs):
    """
    Convenience function for contour detection.

    This function accounts for the OpenCV version supplied and executes
    accordingly.

    Parameters
    ----------
    img : ndarray
        Grayscale image to perform contour detection on.
    *args : tuple
        Additional arguments `cv2.findContours` will require.
    *kwargs : tuple
        Additional keyword arguments `cv2.findContours` will require.

    Returns
    -------
    contours : ndarray
        An array of contours detected.
    hierarchy : ndarray
        Corresponding hierarchy information to the contours detected.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale or has greater than 2 channels.

    """
    _channel_check(img, 2)

    if cv_vers == 4:
        contours, hierarchy = cv2.findContours(img, *args, **kwargs)
    else:
        _, contours, hierarchy = cv2.findContours(img, *args, **kwargs)
    return contours, hierarchy


def fit_circles(contours):
    """
    Fit minimum enclosing circles to contour(s).

    Parameters
    ----------
    contours : ndarray or list
        Contour(s) to fit circles to.

    Returns
    -------
    circles : ndarray or list
        An array or list corresponding to dimensions to circles fitted.

    """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        ret = [cv2.minEnclosingCircle(c) for c in contours]
    else:
        ret = cv2.minEnclosingCircle(contours)
    return ret


def fit_rectangles(contours, rotate=False):
    """
    Fit bounding boxes to contour(s).

    Parameters
    ----------
    contours : ndarray or list
        Contour(s) to fit circles to.
    rotate : bool
        Whether to fit rotated bounding boxes, default is False.

    Returns
    -------
    rectangles : ndarray or list
        An array or list corresponding to dimensions to bounding boxes fitted.

    """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        if rotate:
            ret = [cv2.minAreaRect(c) for c in contours]
        else:
            ret = [cv2.boundingRect(c) for c in contours]
    else:
        if rotate:
            ret = cv2.minAreaRect(contours)
        else:
            ret = cv2.boundingRect(contours)
    return ret


def fit_ellipses(contours):
    """
    Fit ellipses to contour(s).

    Parameters
    ----------
    contours : ndarray or list
        Contour(s) to fit ellipses to.

    Returns
    -------
    ellipses : ndarray or list
        An array or list corresponding to dimensions to ellipses fitted.

    """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        ret = [cv2.fitEllipse(c) for c in contours]
    else:
        ret = cv2.fitEllipse(contours)
    return ret


def draw_contours(img, contours, *args, **kwargs):
    """
    Convenience function for drawing contour(s) on an image.

    Parameters
    ----------
    img : ndarray
        Image to draw contours on.
    contours : ndarray or list
        Contour(s) to draw on the supplied image.
    *args : tuple
        Additional arguments `cv2.drawContours` will require.
    *kwargs : tuple
        Additional keyword arguments `cv2.drawContours` will require.

    See Also
    --------
    draw_rectangles
    draw_circles
    draw_ellipses

    """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        for c in contours:
            cv2.drawContours(img, [c], *args, **kwargs)
    else:
        cv2.drawContours(img, [contours], *args, **kwargs)


def draw_rectangles(img, dims, *args, **kwargs):
    """
    Convenience function for drawing rectangle(s) on an image.

    Parameters
    ----------
    img : ndarray
        Image to draw contours on.
    dims : tuple or list
        Rectangle(s) to draw on the supplied image. Note these should be
        supplied in (x,y,w,h) format.
    *args : tuple
        Additional arguments `cv2.rectangle` will require.
    *kwargs : tuple
        Additional keyword arguments `cv2.rectangle` will require.

    See Also
    --------
    draw_contours
    draw_circles
    draw_ellipses

    """
    def _draw(rect):
        x,y,w,h = rect
        cv2.rectangle(img, (x, y), (x + w, y + h), *args, **kwargs)

    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        for r in dims: _draw(r)
    else:
        _draw(dims)

def draw_circles(img, dims, *args, **kwargs):
    """
    Convenience function for drawing circle(s) on an image.

    Parameters
    ----------
    img : ndarray
        Image to draw contours on.
    dims : tuple or list
        Circle(s) to draw on the supplied image. Note these should be
        supplied in ((x,y),r) format.
    *args : tuple
        Additional arguments `cv2.circle` will require.
    *kwargs : tuple
        Additional keyword arguments `cv2.circle` will require.

    See Also
    --------
    draw_contours
    draw_rectangles
    draw_ellipses

    """
    def _draw(circ):
        (x,y),r = circ
        x,y,r = map(int, (x,y,r))
        cv2.circle(img, (x,y), r, *args, **kwargs)

    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        for c in dims: _draw(c)
    else:
        _draw(c)


def draw_ellipses(img, dims, *args, **kwargs):
    """
    Convenience function for drawing ellipse(s) on an image.

    Parameters
    ----------
    img : ndarray
        Image to draw contours on.
    dims : tuple or list
        Ellipse(s) to draw on the supplied image. Note these should be
        supplied in ((x,y),(w,h),a) format.
    *args : tuple
        Additional arguments `cv2.ellipse` will require.
    *kwargs : tuple
        Additional keyword arguments `cv2.ellipse` will require.

    See Also
    --------
    draw_contours
    draw_rectangles
    draw_circles

    """

    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        for c in dims:
            cv2.ellipse(img, c, *args, **kwargs)
    else:
        cv2.ellipse(img, dims, *args, **kwargs)


def gray(frame):
    """
    Convert a frame to grayscale.

    Parameters
    ----------
    frame : ndarray
        Frame to grayscale.

    Returns
    -------
    frame : ndarray
        Grayscale frame.

    Raises
    ------
    ValueError
        If the supplied image is not of BGR format.

    See Also
    --------
    bgr
    hsv

    """
    _channel_check(frame, 3)
    return cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)


def bgr(frame):
    """
    Convert a frame to BGR format.

    Parameters
    ----------
    frame : ndarray
        Frame to convert.

    Returns
    -------
    frame : ndarray
        Frame of BGR format.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    gray
    hsv

    """
    _channel_check(frame, 2)
    return cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)


def hsv(frame):
    """
    Convert a frame to HSV format.

    Parameters
    ----------
    frame : ndarray
        Frame to convert.

    Returns
    -------
    frame : ndarray
        Frame of HSV format.

    Raises
    ------
    ValueError
        If the supplied image is not of BGR format.

    See Also
    --------
    gray
    bgr

    """
    _channel_check(frame, 3)
    return cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)


class Mask:
    """
    Convenience class for performing segmentation to a mask.

    ``Mask`` enables the creation of a callable that will perform
    ``bitwise_and`` on a given frame with the mask supplied at initiation.
    This can be useful for performing the same operation on a series of frames.

    Parameters
    ----------
    mask : ndarray
        Mask to segment frame(s) to.

    Returns
    -------
    mask : Mask
        Class object with ``__call__`` method that performs ``bitwise_and``
        with the mask provided at initiation.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    shrink
    rect_mask
    circle_mask
    contour_mask

    """

    def __init__(self, mask):
        _channel_check(mask, 2)
        self.mask = mask

    def __call__(self, frame) -> "Mask":
        """
        Mask a frame using the pre-defined mask.

        Parameters
        ----------
        frame : ndarray.
            Frame to mask.

        Returns
        -------
        masked_frame : ndarray
            Masked frame.

        Raises
        ------
        ValueError
            If the supplied image is not grayscale.

        """
        _channel_check(frame, 2)
        return cv2.bitwise_and(frame, frame, mask=self.mask)


def cast_contours(contours, x, y):
    """
    Convenience function to cast contour(s) to a given x,y.

    Parameters
    ----------
    contours : list or ndarray
        Contour(s) to cast.
    x : int
        x position to cast to.
    y : int
        y position to cast to.

    Returns
    -------
    contours : list of ndarray
        Contours casted to new coordinates.

    See Also
    --------
    find_contours
    draw_contours

    """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        for i, c in enumerate(contours):
            c[:, 0, 0] += x
            c[:, 0, 1] += y
            contours[i] = c
    else:
        contours[:, 0, 0] += x
        contours[:, 0, 1] += y
    return contours


def shrink(img, by=50):
    """
    Mask an image to a new roi.

    Parameters
    ----------
    img : ndarray
        Grayscale image to reduce the roi for.
    by : int
        Number of pixels to reduce roi by, default is 50.

    Returns
    -------
    img : ndarray
        Image with reduced roi.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    Notes
    -----
    This function will shrink the roi by a uniform amount on all four
    sides of an image, and not to a specific region within the image.

    See Also
    --------
    Mask
    rect_mask
    circle_mask
    contouer_mask

    """
    _channel_check(img, 2)

    mask = np.zeros_like(img)
    h, w = img.shape
    mask[by:h-by, by:w-by] = 1
    img = cv2.bitwise_and(img, img, mask=mask)
    return img


def rect_mask(img, dims):
    """
    Create a rectangular mask.

    Parameters
    ----------
    img : ndarray
        Grayscale image to produce the rectangular mask for.
    dims : tuple or list
        Rectangle(s) to base mask on. Note these should be
        supplied in (x,y,w,h) format.

    Returns
    -------
    mask : ndarray
        Rectangular mask.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    Mask
    shrink
    circle_mask
    contour_mask

    """
    _channel_check(img, 2)

    rect_mask_ = np.zeros_like(img)
    draw_rectangles(rect_mask_, dims, 255, -1)
    return rect_mask_


def circle_mask(img, dims):
    """
    Create a circular mask.

    Parameters
    ----------
    img : ndarray
        Grayscale image to produce circular mask for.
    dims : tuple or list
        Circle(s) to base mask on. Note these should be
        supplied in ((x,y),r) format.

    Returns
    -------
    mask : ndarray
        Circular mask.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    Mask
    shrink
    rect_mask
    contour_mask

    """
    _channel_check(img, 2)

    circle_mask_ = np.zeros_like(img)
    draw_circles(circle_mask_, dims, 255, -1)
    return circle_mask_

def ellipse_mask(img, dims):
    """
    Create an ellipse mask.

    Parameters
    ----------
    img : ndarray
        Grayscale image to produce elliptical mask for.
    dims : tuple or list
        Ellipse(s) to base mask on. Note these should be
        supplied in ((x,y),(w,h),a) format.

    Returns
    -------
    mask : ndarray
        Elliptical mask.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    Mask
    shrink
    rect_mask
    circle_mask
    contour_mask

    """
    _channel_check(img, 2)

    ell_mask_ = np.zeros_like(img)
    draw_ellipses(ell_mask_, dims, 255, -1)
    return ell_mask_

def contour_mask(img, contours):
    """
    Create a contour mask at the dimensions of contour(s).

    Parameters
    ----------
    img : ndarray
        Grayscale image to produce circular mask for.
    contours : list or ndarray
        Contour(s) to create mask with.

    Returns
    -------
    mask : ndarray
        Contour mask at the dimensions of the contour(s) supplied.

    Raises
    ------
    ValueError
        If the supplied image is not grayscale.

    See Also
    --------
    Mask
    shrink
    rect_mask
    circle_mask

    """
    _channel_check(img, 2)

    cnt_mask = np.zeros_like(img)
    draw_contours(cnt_mask, contours, -1, 255, -1)
    return cnt_mask


# Contour filters ----------------------------------------------------------------------------


def _contours_area(contours):
    """Convenience function for computing the area of contour(s). """
    if isinstance(contours, list) or isinstance(contours, np.ndarray) or isinstance(contours, tuple):
        areas = [cv2.contourArea(c) for c in contours]
    else:
        areas = cv2.contourArea(contours)
    return areas


def smallest(contours):
    """
    Contour filter that returns the smallest contour by area.

    Parameters
    ----------
    contours : list or ndarray
        Contour(s) to filter.

    Returns
    -------
    contour : ndarray
        Smallest contour by area.

    See Also
    --------
    largest
    parents
    Area
    Eccentricity
    Solidity

    """
    areas = _contours_area(contours)
    return contours[np.argmin(areas)]


def largest(contours):
    """
    Contour filter that returns the largest contour by area.

    Parameters
    ----------
    contours : list or ndarray
        Contour(s) to filter.

    Returns
    -------
    contour : ndarray
        Largest contour by area.

    See Also
    --------
    smallest
    parents
    Area
    Eccentricity
    Solidity

    """
    areas = _contours_area(contours)
    return contours[np.argmax(areas)]


def parents(contours, hierarchy):
    """
    Contour filter that returns only parent contours.

    Parameters
    ----------
    contours : list or ndarray
        Contour(s) to filter.
    hierarchy : ndarray
        Corresponding hierarchy information for contours.

    Returns
    -------
    contour : ndarray
        All parent contours.

    See Also
    --------
    smallest
    largest
    Area
    Eccentricity
    Solidity

    """
    parentLvl = hierarchy[0, :, 3].tolist()
    return [contours[i] for i, e in enumerate(parentLvl) if e == -1]


class _ByLimit:
    """
    Filter values based on pre-defined limits.

    """

    def __init__(self, min=None, max=None):
        if min is None and max is None:
            raise ValueError("Limit(s) required for filtering contours.")

        self.min = min
        self.max = max

    def __call__(self, x):
        """
        Filter x based on  a lower or upper limit, or both. If x is between either
        of the pre-defined limits then it will be returned.

        """
        min, max = (self.min, self.max)
        if min and max:
            if x >= min and x <= max:
                return x
        else:
            if min and not max:
                if x >= min:
                    return x
            elif max and not min:
                if x <= max:
                    return x


class Area:
    """
    Filter contours by area based on a lower or upper limit, or both.

    Parameters
    ----------
    min : int or float
        Lower limit to filter contour areas.
    max: int or float
        Upper limit to filter contour areas.

    Returns
    -------
    filter : Area
        Class object with ``__call__`` method that filters contours based on
        the pre-defined area limits provided at initiation.

    See Also
    --------
    smallest
    largest
    parents
    Eccentricity
    Solidity

    """

    def __init__(self, min=None, max=None):
        self._filter = _ByLimit(min, max)

    def __call__(self, contours) -> "Area":
        """
        Filter contours by area based on pre-defined limits.

        Parameters
        ----------
        contours : list
            List of contours to filter.

        Returns
        -------
        contours : list
            Filtered contours.

        """
        areas = _contours_area(contours)
        _filtered = []

        for i, a in enumerate(areas):
            if self._filter(a):
                _filtered.append(contours[i])

        return _filtered


class Eccentricity:
    """
    Filter contours by eccentricity based on a lower or upper limit, or both.

    Parameters
    ----------
    min : int or float
        Lower eccentricity limit to filter contours.
    max: int or float
        Upper eccentricity limit to filter contours.

    Returns
    -------
    filter : Eccentricity
        Class object with ``__call__`` method that filters contours based on
        the pre-defined eccentricity limits provided at initiation.

    See Also
    --------
    smallest
    largest
    parents
    Area
    Eccentricity

    """

    def __init__(self, min=None, max=None):
        self._filter = _ByLimit(min, max)

    def _eccentricity(self, contours):
        """Convenience method for computing the eccentricity of contours. """
        for i, c in enumerate(contours):
            if c.shape[0] >= 5:  # filter to this size for below calculations
                center, axes, orientation = cv2.fitEllipse(c)
                major, minor = (max(axes), min(axes))
                eccentricity = np.sqrt(1 - (minor / major) ** 2)
                yield (i, eccentricity)

    def __call__(self, contours) -> "Eccentricity":
        """
        Filter contours by eccentricity based on pre-defined limits.

        Parameters
        ----------
        contours : list
            List of contours to filter.

        Returns
        -------
        contours : list
            Filtered contours.

        """
        _filtered = []
        for i, e in self._eccentricity(contours):
            if self._filter(e):
                _filtered.append(contours[i])

        return _filtered


class Solidity:
    """
    Filter contours by solidity based on a lower or upper limit, or both.

    Parameters
    ----------
    min : int or float
        Lower solidity limit to filter contours.
    max: int or float
        Upper solidity limit to filter contours.

    Returns
    -------
    filter : Solidity
        Class object with ``__call__`` method that filters contours based on
        the pre-defined solidity limits provided at initiation.

    Notes
    -----
    Note that here the solidity of a contour is based on its area relative
    to its corresponding convex hull.

    See Also
    --------
    smallest
    largest
    parents
    Area
    Eccentricity

    """

    def __init__(self, min=None, max=None):
        self._filter = _ByLimit(min, max)

    def _solidity(self, contours):
        """Convenience method for computing the solidity of contours. """
        for i, c in enumerate(contours):
            c_area = cv2.contourArea(c)
            hull = cv2.convexHull(c)
            hullarea = cv2.contourArea(hull)

            try:
                yield (i, c_area / hullarea)
            except ZeroDivisionError:
                pass

    def __call__(self, contours) -> "Solidity":
        """
        Filter contours by solidity based on pre-defined limits.

        Parameters
        ----------
        contours : list
            List of contours to filter.

        Returns
        -------
        contours : list
            Filtered contours.

        """
        _filtered = []
        for i, s in self._solidity(contours):
            if self._filter(s):
                _filtered.append(contours[i])

        return _filtered



from tqdm import tqdm
import cv2
import numpy as np
import glob
import math
import copy
from more_itertools import prepend
from natsort import natsorted, ns
import warnings

from vuba import ops


def take_first(it):
    """
    Retrieve the first value from an iterable object.

    Parameters
    ----------
    it : iterable
        Iterable object to retrieve the first value from.

    Returns
    -------
    first : number or string or ndarray or None
        First value from the iterable object. Will be
        None if ``StopIteration`` is raised.

    """

    it = iter(it)
    try:
        return next(it)
    except StopIteration:
        return None


def fourcc_to_string(fourcc):
    """
    Convert fourcc integer code into codec string.

    Parameters
    ----------
    fourcc : int
        Fourcc integer code.

    Returns
    -------
    codec : str
        Codec string corresponding to fourcc code.

    """
    char1 = str(chr(fourcc & 255))
    char2 = str(chr((fourcc >> 8) & 255))
    char3 = str(chr((fourcc >> 16) & 255))
    char4 = str(chr((fourcc >> 24) & 255))

    return char1 + char2 + char3 + char4


def open_video(video):
    """
    Convenience function for creating a ``Video`` instance.

    Parameters
    ----------
    video : str or cv2.VideoCapture or list.
        Full filename or VideoCapture object to a video (e.g. AVI or MP4),
        or a glob string or series of filenames to a series of individual
        images.

    Returns
    -------
    video : Video
        Instance of ``Video``.
    release : bool
        True if footage was not an instance to ``Video``, else False.

    """
    release = False
    if not isinstance(video, Video):
        video = Video(video)
        release = True

    return video, release


class Frames:
    """
    Container for frames that is used in ``Video.read``.

    Parameters
    ----------
    reader : callable
        Frame reader method supplied from ``Video`` class.
    start : int
        Index of first frame.
    stop : int
        Index of last frame.
    step : int
        Step size.
    low_memory : bool
        Whether to import frames into memory.
    grayscale : bool
        Whether to convert frames to grayscale.

    Returns
    -------
    frames : Frames
        Container for frames.

    Notes
    -----
    Frames are imported into memory at initiation if low_memory=False.

    See Also
    --------
    Video

    """

    def __init__(self, reader, start, stop, step, low_memory, grayscale):
        self._reader = reader
        self.idxs = (start, stop, step)
        self.grayscale = grayscale
        self.in_mem = False
        self._roi = None

        if not low_memory:
            self.import_to_ndarray()

    def import_to_ndarray(self) -> "Frames":
        """
        Import the declared frames into a contiguous numpy array.

        """
        gen_fr = self._reader(*self.idxs, self.grayscale)
        first = take_first(gen_fr)
        self.ndarray = np.ascontiguousarray(
            np.empty((len(self), *first.shape), dtype="uint8")
        )
        gen_fr = prepend(first, gen_fr)

        # print("Importing frames into memory...")
        # pg = tqdm(total=len(self))
        for i, frame in enumerate(gen_fr):
            self.ndarray[i, :] = frame[:]
            # pg.update(1)
        # pg.close()

        self.in_mem = True

    def __len__(self) -> "Frames":
        """
        Retrieve the length of the provided frames without having to iterate
        across them.

        Returns
        -------
        length : int
            Length of the provided frames.

        """
        (start, stop, step) = self.idxs
        return math.floor((stop - start) / step)

    def __iter__(self) -> "Frames":
        """
        Retrieve the frames declared at initiation.

        Returns
        -------
        frames : generator
            Generator that supplies frames.

        """
        if self.in_mem:
            for frame in self.ndarray:
                yield frame
        else:
            for frame in self._reader(*self.idxs, self.grayscale):
                yield frame


class Video:
    """
    Wrapper around various image readers that provides a simple API
    to achieve the same functions regardless of format.

    Parameters
    ----------
    footage : str or ``cv2.VideoCapture`` or list.
        Full filename or VideoCapture object to a video (e.g. AVI or MP4),
        or a glob string or series of filenames to a series of individual
        images. Note that series of all individual filenames will be sorted
        prior to being read if a glob string is supplied.

    Returns
    -------
    video : Video
        Video handler for footage supplied.

    See Also
    --------
    open_video
    Frames
    Writer

    Examples
    --------
    For example usage of this handler please see example scripts located
    at ``examples/reading``.

    """

    def __init__(self, footage):
        def open_videocv(footage):
            self.video = footage
            self.video_release = False
            if not isinstance(self.video, cv2.VideoCapture):
                self.video = cv2.VideoCapture(self.video)
                self.video_release = True
            self.filenames = None

        if isinstance(footage, list):
            self.filenames = footage
            self.video = self.video_release = None
        else:
            try:
                # Somewhat hacky but provides consistent behaviour
                if "*" in footage:
                    files = glob.glob(footage)
                    self.filenames = natsorted(files, alg=ns.IGNORECASE)
                    self.video = self.video_release = None
                else:
                    open_videocv(footage)
            except TypeError:
                open_videocv(footage)
            except:
                raise

        self._grab_info()

    def __len__(self) -> "Video":
        """
        Retrieve the length of the provided footage without having to iterate
        across it.

        Returns
        -------
        length : int
            Length of the provided footage.

        """
        if self.video:
            return int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))
        else:
            return len(self.filenames)

    def _grab_info(self):
        """
        Grab summary footage information.

        This method is called on initiation of ``Video``.

        """
        if self.video:
            self.fps = round(self.video.get(cv2.CAP_PROP_FPS))

            width = self.video.get(cv2.CAP_PROP_FRAME_WIDTH)
            height = self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)
            self.resolution = (self.width, self.height) = tuple(
                map(round, (width, height))
            )

            self.fourcc_code = int(self.video.get(cv2.CAP_PROP_FOURCC))
            self.codec = fourcc_to_string(self.fourcc_code)
        else:
            first_frame = self.read(0)
            width = first_frame.shape[1]
            height = first_frame.shape[0]
            self.resolution = (self.width, self.height) = tuple(
                map(round, (width, height))
            )

            self.fps = self.codec = self.fourcc_code = None

    def read(self, *args, **kwargs) -> "Video":
        """
        Read single or multiple frames from the provided footage.

        Note that reading multiple frames follows slice behaviour,
        whereby there is a 'start', 'stop' and 'step'. However steps will
        only give a performance uplift for reading images stored as
        individual files (e.g. multiple pngs). If you are reading from
        videos, the read time will be the same as if you were reading
        all the frames between 'start' and 'stop'. This is because it is
        faster to pass on frames that don't match the step size, than it is
        to repeatedly locate the correct frames by index in the video file.

        Parameters
        ----------
        * For single frames:
            index : int
                Index of frame in footage to read.
            grayscale : bool
                Whether to convert frame to grayscale.

        * For multiple frames:
            start : int
                Index of frame to start reading from.
            stop : int
                Index of frame to stop reading at.
            step : int
                Step size.
            low_memory : bool
                Whether to import frames into RAM, default is True i.e. not to.
            grayscale
                Whether to convert frames to grayscale.

        Returns
        -------
        For single frames:
            frame : ndarray
                Frame at the given index.

        For multiple frames:
            frames : Frames
                Container that will either supply the frames from
                an ndarray or from a generator. Note that this
                container contains both a len and iter method. For the
                latter, a new generator is created upon calling the
                method if the frames have not been imported into memory.
                This is to maintain implementation parity with the
                in-memory container.

        Raises
        ------
        ValueError
            If the supplied indices are not within the available range of the
            selected footage, or if there is a missing required argument.

        Examples
        --------
        For example usage of this method please see example scripts located
        at ``examples/reading``.

        """
        try:
            frame = self._read_single(*args, **kwargs)
            return frame
        except TypeError:
            args = self._prep_args_kwargs(*args, **kwargs)
            frames = Frames(self._read_multi, *args)
            return frames
        except:
            raise

    def _prep_args_kwargs(
        self, start=None, stop=None, step=1, low_memory=True, grayscale=False
    ):
        """
        Prep any args and kwargs supplied from Video.read().

        """

        def check_indexes(index):
            if index is not None:
                if index < 0 or index > len(self):
                    raise IndexError(
                        "Indices out of available range for provided footage."
                    )
            else:
                raise ValueError(f"Missing required argument.")

        check_indexes(start)
        check_indexes(stop)

        return (start, stop, step, low_memory, grayscale)

    def _read_single(self, index, grayscale=False):
        """
        Read a single frame at a given location in the requested footage.

        """
        if self.video:
            self.video.set(1, index)
            success, frame = self.video.read()
        else:
            frame = cv2.imread(self.filenames[index])

        if grayscale:
            frame = ops.gray(frame)

        return frame

    def _read_multi(self, start, stop, step, grayscale):
        """
        Read multiple frames from the requested footage at the given indices.

        """
        if self.video:
            self.video.set(1, start)

            if step:
                if step > 1:
                    # Floor is used here to stop out-of-range indices being created
                    number = math.floor((stop - start) / step)
                    frames_to_yield = [round((step * n) + start) for n in range(number)]
                    steps = True
                else:
                    steps = False

            for fr in range(start, stop):
                success, frame = self.video.read()
                if not success:
                    break

                if grayscale:
                    frame = ops.gray(frame)

                if steps:
                    if fr in frames_to_yield:
                        yield frame
                    else:
                        pass
                else:
                    yield frame
        else:
            for fn in self.filenames[slice(start, stop, step)]:
                frame = cv2.imread(fn)
                if grayscale:
                    frame = ops.gray(frame)
                yield frame

    def close(self) -> "Video":
        """
        Close/teardown attached video handlers.

        """
        if self.video:
            self.video.release()
        elif self.filenames:
            self.filenames = None


class Writer:
    """
    Create an encoder for exporting frames at a given output.

    Regardless of input format, this encoder can encode footage to both individual
    images and to video files. For encoding to individual images, a sequence of
    filenames needs to be supplied that will match the length of the number of frames
    that need to be encoded. Conversely, for encoding to a video file, a filename to
    the video file needs to be provided.

    Parameters
    ----------
    output : str or list
        Output to export frames.
    footage : Video
        Instance of Video to create writer based on. If this is
        not supplied then you must supply all encoder specific arguments.
    fps : float
        Framerate to export footage at, default is the fps of
        the video supplied. Note that this argument must be supplied
        if working with individual images.
    resolution : tuple
        Width and height to export footage at (both must be supplied
        as integers). Default is the resolution of the footage supplied.
    codec : str
        Codec to encode footage with. Default is to encode with MJPG
        codec or that of the Video instance supplied (if any).
    grayscale : bool
        Whether to export footage as grayscale. Default is False.

    Returns
    -------
    encoder : Writer
        Encoder to export frames to a given output.

    Raises
    ------
    ValueError
        If a frame-rate is not supplied when encoding to a movie using
        individual images.

    See Also
    --------
    Video

    Examples
    --------
    For example usage of this handler please see example scripts located
    at ``examples/writing``.

    """

    def __init__(
        self,
        output,
        footage=None,
        fps=None,
        resolution=None,
        codec=None,
        grayscale=False,
    ):
        self.footage_release = None
        if footage:
            footage, self.footage_release = open_video(footage)

        self.is_color = True
        if grayscale:
            self.is_color = False

        self.resolution = resolution
        if not resolution:
            self.resolution = footage.resolution

        self.encoder_release = False
        if isinstance(output, str):
            if not codec:
                if footage:
                    if footage.video:
                        codec = footage.codec
                    else:
                        codec = "MJPG"
                else:
                    codec = "MJPG"

            fourcc = cv2.VideoWriter_fourcc(*codec)

            if not fps:
                if footage:
                    if footage.fps:
                        fps = footage.fps
                    else:
                        raise ValueError(
                            "When working with individual images you must supply a frame-rate to export footage at."
                        )
                else:
                    raise ValueError(
                        "Integer must be supplied to fps when an instance of vuba.Video is not supplied."
                    )

            self.encoder = cv2.VideoWriter(
                output, fourcc, fps, self.resolution, self.is_color
            )
            self.encoder_release = True
        elif isinstance(output, list):
            self.encoder = iter(output)
            self.fps = self.fourcc = None

    def write(self, frame) -> "Writer":
        """
        Write a frame using the declared encoder.

        Parameters
        ----------
        frame : ndarray
            Frame to export.

        Warns
        -----
        UserWarning
            If the supplied frame is not of the correct resolution or colour space,
            and to notify the user that the frame has been correctly converted so
            encoding is successful.

        Notes
        -----
        For encoding to video, note that if the frames supplied for encoding are not at
        the same resolution as that set for the encoder upon initiation, the frames will
        be resized accordingly. Also note that frames will be converted to the colour
        space set at initiation if they are not the same. These features are built-in so
        that a complete movie is always created, rather than an empty container which
        is not of much use for debugging a failed component of a script.

        Examples
        --------
        For example usage of this method please see example scripts located
        at ``examples/writing``.

        """
        if isinstance(self.encoder, cv2.VideoWriter):
            size = (frame.shape[1], frame.shape[0])
            if size != self.resolution:
                frame = cv2.resize(frame, self.resolution)
                warnings.warn(
                    f"Frame has been resized to ensure successful encoding. Supplied frame vs encoder resolution: {size} , {self.resolution}",
                    UserWarning,
                )

            channels = len(frame.shape)
            if channels > 2 and not self.is_color:
                frame = ops.gray(frame)
                warnings.warn(
                    f"Frame has been grayscaled to ensure successful encoding. Supplied frame vs encoder format: {channels} , {2, self.is_color}",
                    UserWarning,
                )
            elif channels == 2 and self.is_color:
                frame = ops.bgr(frame)
                warnings.warn(
                    f"Frame has been converted to BGR format to ensure successful encoding. Supplied frame vs encoder format: {channels} , {3, self.is_color}",
                    UserWarning,
                )
            self.encoder.write(frame)
        else:
            cv2.imwrite(next(self.encoder), frame)

    def close(self) -> "Writer":
        """
        Close/teardown declared encoders.

        """
        if self.encoder_release:
            self.encoder.release()
        if self.footage_release:
            self.footage.close()



from setuptools import setup
import codecs
import os.path
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

# For reading in the version string without importing the package
# Ref: https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="vuba",
    version=get_version("vuba/__init__.py"),
    description="An easy to use library for constructing OpenCV HighGUI interfaces.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EmbryoPhenomics/vuba",
    author="Ziad Ibbini",
    author_email="ziad.ibbini@students.plymouth.ac.uk",
    license="MIT",
    packages=["vuba"],
    python_requires=">=3.5, <4",
    install_requires=[
        "numpy",
        "tqdm",
        "opencv-python",
        "more_itertools",
        "dataclasses",
        "natsort",
    ],
    project_urls={"Source": "https://github.com/EmbryoPhenomics/vuba/tree/main/vuba"},
)



import functools
import cv2
from typing import Callable, Iterable
from dataclasses import dataclass
import numpy as np

from vuba import imio


@dataclass
class TrackbarMethod:
    """
    Container for a trackbar method and it's associated variables.

    Parameters
    ----------
    id : str
        Identification string of trackbar added in ``BaseGUI.trackbar``.
    min : int
        Minimum limit of trackbar.
    max : int
        Maximum limit of trackbar.
    method : callable
        Trackbar callback as specified in ``BaseGUI.trackbar``.
    current : int
        Current trackbar value.

    Returns
    -------
    container : dataclass
        Mutable container for a trackbar and it's associated variables.

    """

    __slots__ = ["id", "min", "max", "method", "current"]
    id: str  #: Identification string of trackbar added in ``BaseGUI.trackbar``.
    min: int  #: Minimum limit of trackbar.
    max: int  #: Maximum limit of trackbar.
    method: Callable  #: Trackbar callback as specified in ``BaseGUI.trackbar``.
    current: int  #: Current trackbar value.


class BaseGUI:
    """
    The base constructor class for creating HighGUI interfaces.

    Typical usage of this class is through the supplied wrappers around
    this class, although if those are ill-fit for a given application this
    constructor can be addressed directly.

    To use this constructor with single or multiple frames that are in memory,
    use the ``FrameGUI`` or ``FramesGUI`` classes respectively, Conversely, to use
    this constructor with video files or streams, use the ``VideoGUI`` or ``StreamGUI``
    classes respectively.

    Parameters
    ----------
    title : str
        Title of the interface window.

    Returns
    -------
    gui : BaseGUI
        The newly created gui constructor.

    See Also
    --------
    FrameGUI
    FramesGUI
    VideoGUI
    StreamGUI

    """

    def __init__(self, title):
        self.title = title
        self.trackbars = {}
        cv2.namedWindow(self.title)

    def values(self) -> "BaseGUI":
        """
        Retrieve all current trackbar values from the interface.

        Returns
        -------
        values : dict or None
            All current trackbar values, returns None if no
            trackbars declared.

        See Also
        --------
        BaseGUI.__getitem__
        BaseGUI.__setitem__

        """
        if self.trackbars:
            vals = {}
            for k in self.trackbars:
                vals[k] = self.trackbars[k].current
        else:
            vals = None
        return vals

    def __getitem__(self, key) -> "BaseGUI":
        """
        Access a trackbar's current value.

        Parameters
        ----------
        key : str
            Trackbar identification string provided
            when using ``BaseGUI.trackbar``.

        Returns
        -------
        value : int
            Trackbar's current value.

        See Also
        --------
        BaseGUI.values
        BaseGUI.__setitem__

        """
        return self.trackbars[key].current

    def __setitem__(self, key, val) -> "BaseGUI":
        """
        Change a trackbar's current recorded value.

        Parameters
        ----------
        key : str
            Trackbar identification string provided
            when using ``BaseGUI.trackbar``.
        val : int
            Value to record trackbar as.

        Notes
        -----
        ``BaseGUI.__setitem__`` will not change the trackbar value in
        the OpenCV interface and will only change the value recorded
        in the ``BaseGUI`` class instance.

        See Also
        --------
        BaseGUI.values
        BaseGUI.__getitem__

        """
        self.trackbars[key].current = val

    def method(self, func) -> "BaseGUI":
        """
        Add a main processing function to be executed on every trackbar call.

        All trackbar functions should call the function that is wrapped
        in this decorator. As such changes made to a single trackbar, will
        propagate to this main processing function and the changes will be
        returned to the named window.

        Parameters
        ----------
        func : callable
            Main processing function for the interface.

        See Also
        --------
        BaseGUI.trackbar

        Examples
        --------
        To demonstrate this method's usage as a decorator we can construct
        a simple binary threshold viewer for a random binary image:

        >>> import vuba
        >>> import numpy as np
        >>> import cv2
        >>> img = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
        >>> gui = vuba.BaseGUI('Binary threshold viewer')
        >>> @gui.method
        >>> def threshold(gui):
        ...     frame = img.copy()
        ...     thresh_val = gui['thresh_val']
        ...     _, thresh = cv2.threshold(frame, thresh_val, 255, cv2.THRESH_BINARY)
        ...     return thresh
        >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
        >>> gui.run()

        Note that this method does not have to be used as a decorator:

        >>> gui = vuba.BaseGUI('Binary threshold viewer')
        >>> def threshold(gui):
        ...     frame = img.copy()
        ...     thresh_val = gui['thresh_val']
        ...     _, thresh = cv2.threshold(frame, thresh_val, 255, cv2.THRESH_BINARY)
        ...     return thresh
        >>> gui.method(threshold)
        >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
        >>> gui.run()

        Also note that since we are not doing any additional processing on
        each trackbar call in the above two examples, we can supply None to the
        ``BaseGUI.trackbar`` decorator. This will default to using a simple callback
        that will pass the trackbar values to the main processing method
        and display the result in the interactive window.

        """

        @functools.wraps(func)
        def wrap_to_proc():
            img = func(self)
            return img

        self.process = wrap_to_proc
        return wrap_to_proc

    @staticmethod
    def _basic_trackbar_callback(gui, id, val):
        """
        Basic trackbar callback to be used when ``BaseGUI.trackbar`` is not supplied a function.

        """
        gui[id] = val
        img = gui.process()
        cv2.imshow(gui.title, img)

    def trackbar(self, name, id, min, max) -> "BaseGUI":
        """
        Add a trackbar to the interactive window.

        Parameters
        ----------
        name : str
            Name of trackbar to add.
        id : str
            Id of trackbar to add. This will be the key associated with the trackbar.
        min : int
            Minimum limit of trackbar
        max : int
            Maximum limit of trackbar
        func : callable, optional
            Trackbar callback to be called whenever the trackbar is changed. If
            None, then a basic callback will be used that simply passes the trackbar
            value to the ``BaseGUI.method`` method.

        Notes
        -----
        This method can be called in much the same way as ``BaseGUI.method``, either as
        a decorator or as a typical method call.

        Examples
        --------
        For these examples, we will use the same binary threshold viewer as used in
        ``BaseGUI.method``:

        >>> import vuba
        >>> import numpy as np
        >>> import cv2
        >>> def threshold(gui):
        ...     frame = img.copy()
        ...     thresh_val = gui['thresh_val']
        ...     _, thresh = cv2.threshold(frame, thresh_val, 255, cv2.THRESH_BINARY)
        ...     return thresh
        >>> img = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
        >>> gui = vuba.BaseGUI('Binary threshold viewer')
        >>> gui.method(threshold)

        For applications where there is no additional processing on each trackbar call,
        we can simply supply None to ``BaseGUI.trackbar`` as follows:

        >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
        >>> gui.run()

        However, if we wanted to add some further processing, let's say to exclude
        all odd threshold values, we could handle this logic with a custom callback:

        >>> gui = vuba.BaseGUI('Binary threshold viewer')
        >>> gui.method(threshold)
        >>> @gui.trackbar('Threshold', id='thresh_val', min=0, max=255)
        >>> def even_threshold(gui, val):
        ...     if val%2 == 0:
        ...         gui['thresh_val'] = val
        ...         ret = gui.process()
        ...         cv2.imshow(gui.title, ret)
        >>> gui.run()
        """

        def wrap_to_trackbar(func):
            if not func:
                @functools.wraps(self._basic_trackbar_callback)
                def on_exe(val):
                    return self._basic_trackbar_callback(self, id, val)
            else:
                @functools.wraps(func)
                def on_exe(val):
                    return func(self, val)

            self.trackbars[id] = TrackbarMethod(id, min, max, on_exe, min)
            cv2.createTrackbar(name, self.title, min, max, on_exe)

            return on_exe

        return wrap_to_trackbar

    def run(self) -> "BaseGUI":
        """
        Launch the interface.

        Examples
        --------
        Note that you can access any variables from the class that you
        added/manipulated through the trackbars attribute. This contains
        a dict of the trackbars, with each key containing associated
        variables in a dataclass:

        >>> import vuba
        >>> import numpy as np
        >>> import cv2
        >>> img = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
        >>> gui = vuba.BaseGUI('Binary threshold viewer')
        >>> @gui.method
        >>> def threshold(gui):
        ...     frame = img.copy()
        ...     thresh_val = gui['thresh_val']
        ...     _, thresh = cv2.threshold(frame, thresh_val, 255, cv2.THRESH_BINARY)
        ...     return thresh
        >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
        >>> gui.run()
        >>> name, min, max, method, last_value = gui.trackbars['thresh_val']

        This can be useful for retrieving the ideal parameters for an image analysis
        method after adjusting them in an interface for example.

        """

        # Execute first method to launch the gui
        firstfunc = self.trackbars[[*self.trackbars][0]]
        func = firstfunc.method
        min = firstfunc.min
        func(min)

        cv2.waitKey()
        cv2.destroyAllWindows()


class FrameGUI(BaseGUI):
    """
    Class for creating interfaces for individual image manipulation.

    Parameters
    ----------
    frame : ndarray
        Frame(s) to manipulate within the interface.
    title : str
        Title of the interface window.

    Returns
    -------
    gui : FrameGUI
        Class object for creating the interactive window.

    See Also
    --------
    BaseGUI
    FramesGUI
    VideoGUI
    StreamGUI

    Notes
    -----
    This class does not impose any restrictions on the types/formats of the images
    you supply upon initiation. As such, we can supply a tuple or list of images
    that we want to manipulate at the same time for example (see examples below).

    Examples
    --------
    As in other single image examples, we will use a series of randomly generated
    binary images with a variable binary threshold in the interfaces below:

    >>> import vuba
    >>> import numpy as np
    >>> import cv2
    >>> img = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
    >>> gui = vuba.FrameGUI(img, 'Binary threshold viewer')
    >>> @gui.method
    >>> def threshold(gui):
    ...     frame = gui.frame.copy()
    ...     thresh_val = gui['thresh_val']
    ...     _, thresh = cv2.threshold(frame, thresh_val, 255, cv2.THRESH_BINARY)
    ...     return thresh
    >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
    >>> gui.run()

    As mentioned above, we can supply a series of images for simultaneuous
    manipulation as well:

    >>> img1 = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
    >>> img2 = np.random.randint(low=0, high=255, size=(500, 500), dtype=np.uint8)
    >>> gui = vuba.FrameGUI((img1, img2), 'Binary threshold viewer')
    >>> @gui.method
    >>> def threshold(gui):
    ...     frame1, frame2 = gui.img
    ...     thresh_val = gui['thresh_val']
    ...     _, thresh1 = cv2.threshold(frame1.copy(), thresh_val, 255, cv2.THRESH_BINARY)
    ...     _, thresh2 = cv2.threshold(frame2.copy(), thresh_val, 255, cv2.THRESH_BINARY)
    ...     return np.hstack((thresh1, thresh2))
    >>> gui.trackbar('Threshold', id='thresh_val', min=0, max=255)(None)
    >>> gui.run()

    For an extension of this type of interface with drawing in addition, see
    ``/examples/interfaces/binarythreshod_viewer_with_drawing_video.py``.

    """

    def __init__(self, frame, *args, **kwargs):
        self.frame = frame
        super(FrameGUI, self).__init__(*args, **kwargs)


class FramesGUI(BaseGUI):
    """
    Class for creating interfaces for manipulating a sequence of frames.

    Parameters
    ----------
    frames : list or ndarray
        Images to manipulate within the interface.
    indices : tuple
        Frame indices to limit GUI to, especially useful when analysing long sequences
        of footage. First index will always be interpreted as the minimum index and the 
        second as the maximum index. If None is specified to either limit, then that limit
        will be ignored. Default is for no limits, i.e. (None, None).
    title : str
        Title of the interface window.

    Returns
    -------
    gui : FramesGUI
        Class object for creating the interactive window.

    See Also
    --------
    BaseGUI
    FrameGUI
    VideoGUI
    StreamGUI

    Notes
    -----
    Any gui created with this class will by default have a frame trackbar and
    corresponding callback for retrieving frames from the sequence provided
    at initiation (see ``FramesGUI.read``).

    Examples
    --------
    Since the API for creating an interface is the same regardless of footage format,
    we will demonstrate a basic frame viewer here.

    First, let's create a sequence of binary images that gradually increase
    in their grayscale value:

    >>> import numpy as np
    >>> frames = [np.full((500, 500), i, dtype=np.uint8) for i in range(1, 255)]

    Now let's pass these to our basic frame viewer to view the result:

    >>> import vuba
    >>> import cv2
    >>> gui = vuba.FramesGUI(frames, 'Frame viewer')
    >>> @gui.method
    >>> def blank(gui):
    ...     # Here we are not doing any image processing as we simply
    ...     # want to view the individual frames
    ...     frame = gui.frame.copy()
    ...     return frame
    >>> gui.run()

    """

    def __init__(self, frames, indices=(None, None), *args, **kwargs):
        self.frames = frames
        super(FramesGUI, self).__init__(*args, **kwargs)

        lower_ind, upper_ind = indices
        if not lower_ind:
            lower_ind = 0
        if not upper_ind:
            upper_ind = len(frames)

        self.lower = lower_ind
        self.diff = upper_ind - lower_ind
        self.trackbar("Frames", "frames", 0, self.diff)(self.read)

    @staticmethod
    def read(gui, val) -> "FramesGUI":
        """
        Callback for reading and displaying a frame from the provided frames.

        Parameters
        ----------
        val : int
            Frame index in the requested frames.

        Notes
        -----
        Any image processing in the main method will be executed prior
        to displaying the frame.

        """
        if val >= gui.trackbars["frames"].min and val < gui.trackbars["frames"].max:
            gui["frames"] = val
            gui.frame = gui.frames[int(val + gui.lower)]
            frame_proc = gui.process()
            cv2.imshow(gui.title, frame_proc)


class VideoGUI(BaseGUI):
    """
    Class for creating interfaces for manipulating a sequence of frames.

    Parameters
    ----------
    video : vuba.Video
        Video to manipulate within the interface.
    indices : tuple
        Frame indices to limit GUI to, especially useful when analysing long sequences
        of footage. First index will always be interpreted as the minimum index and the 
        second as the maximum index. If None is specified to either limit, then that limit
        will be ignored. Default is for no limits, i.e. (None, None).
    title : str
        Title of the interface window.

    Returns
    -------
    gui : VideoGUI
        Class object for creating the interactive window.

    See Also
    --------
    BaseGUI
    FrameGUI
    FramesGUI
    StreamGUI

    Notes
    -----
    Any gui created with this class will by default have a frame trackbar and
    corresponding callback for retrieving frames from the video provided
    at initiation (see ``VideoGUI.read``).

    Examples
    --------
    Please see the following example scripts for some typical applications of this
    class:

    * examples/interfaces/frame_viewer.py
    * examples/interfaces/binary_threshold_viewer_with_drawing_video.py

    """

    def __init__(self, video, indices=(None, None), *args, **kwargs):
        self.video, self.release = imio.open_video(video)
        super(VideoGUI, self).__init__(*args, **kwargs)

        lower_ind, upper_ind = indices
        if not lower_ind:
            lower_ind = 0
        if not upper_ind:
            upper_ind = len(video)

        self.lower = lower_ind
        self.diff = upper_ind - lower_ind
        self.trackbar("Frames", "frames", 0, self.diff)(self.read)

    @staticmethod
    def read(gui, val) -> "VideoGUI":
        """
        Callback for reading and displaying a frame from the requested video.

        Parameters
        ----------
        val : int
            Frame index in the requested video.

        Notes
        -----
        Any image processing in the main method will be executed prior
        to displaying the frame.

        """
        if val >= gui.trackbars["frames"].min and val < gui.trackbars["frames"].max:
            gui["frames"] = val
            gui.frame = gui.video.read(int(val + gui.lower), grayscale=False)
            frame_proc = gui.process()
            cv2.imshow(gui.title, frame_proc)

    def run(self) -> "VideoGUI":
        """
        Launch the interactive video interface.

        Notes
        -----
        After running the interface, this function will close/teardown the
        video handler supplied if it was created at initiation.

        """
        try:
            super().run()
        finally:
            if self.release:
                self.video.close()


class StreamGUI(BaseGUI):
    """
    Class for creating interfaces for manipulation of images from video feeds.

    Parameters
    ----------
    stream : Iterable
        An iterator to a camera stream. This can be a generator
        that will continuously pull frames from a capture device
        for example.
    title : str
        Title of the interface window.

    Returns
    -------
    gui : StreamGUI
        Class object for creating the interactive window.

    See Also
    --------
    BaseGUI
    FrameGUI
    FramesGUI
    VideoGUI

    Notes
    -----
    Upon execution of the interface through ``StreamGUI.run``, this interface
    will continuously pull frames from the provided stream and display them in
    the interactive window.

    Examples
    --------
    Please see the following example scripts for some typical applications of this
    class:

    * examples/interfaces/binary_threshold_viewer_with_drawing_camera.py

    """

    def __init__(self, stream, *args, **kwargs):
        self.stream = stream
        super(StreamGUI, self).__init__(*args, **kwargs)

    def run(self) -> "StreamGUI":
        """
        Launch the interactive window.

        Notes
        -----
        Whilst this class overrides ``BaseGUI.run``, exiting the interface is the same
        as other wrappers in that any key press will cause the interface to close.

        If you provide an iterable that has a limit, the interface created will exit
        once it has reached the end of the iterable. As such, we would recommend
        using ``FramesGUI`` when using limited sequences of footage rather streams
        or feeds.

        """
        while True:
            try:
                self.frame = next(self.stream)
            except StopIteration:
                break
            except:
                raise

            frame_proc = self.process()
            cv2.imshow(self.title, frame_proc)

            k = cv2.waitKey(1)
            if k > 0:
                break
